__author__ = 'tomlun'

import pyglet
from pyglet.gl import *


class AnimatedSprite(pyglet.sprite.Sprite):

    def __init__(self, img, rows, columns, duration, batch=None, group=None):
        # create an ImageGrid from the spritesheet
        seq = pyglet.image.ImageGrid(img, rows, columns)
        # create an animation from the ImageGrid
        animation = pyglet.image.Animation.from_image_sequence(seq, duration)
        # pass the animation to the pyglet sprite
        pyglet.sprite.Sprite.__init__(self, animation, batch=batch, group=group)
        # unschedule the animation method so that we will have to start the animation by calling play()
        pyglet.clock.unschedule(self._animate)
        # create an temporary frame range
        self.current_animation = (0, 0)
        self._stopped = True
        self._loop = False

    def _reset_frame_index(self):
        # reset the frame index to the start of the frame range
        self._frame_index = self.current_animation[0]

    def _animate(self, dt):
        # check if we have reached the last frame of
        if self._frame_index > self.current_animation[1]:
            self._reset_frame_index()
            if not self._loop:
                # set the texture to the first frame of the current animation
                frame = self._animation.frames[self._frame_index]
                self._set_texture(frame.image.get_texture())
                # stop animation
                self._stopped = True
                self.dispatch_event('on_animation_end')
                return

        # get the current frame
        frame = self._animation.frames[self._frame_index]
        # set the new texture
        self._set_texture(frame.image.get_texture())

        # increase the frame index
        self._frame_index += 1

        # make sure the duration is set and schedule the next animation for the next frame
        if frame.duration:
            pyglet.clock.schedule_once(self._animate, frame.duration)
        else:
            self._stopped = True
            self.dispatch_event('on_animation_end')

    def play(self, start, end, loop=False):
        if self._stopped:
            # save the new range
            self.current_animation = (start, end)
            self._reset_frame_index()
            # should the animation loop?
            self._loop = loop
            # get the first frame
            frame = self._animation.frames[self._frame_index]
            # start the animation
            self._animate(frame.duration)
            self._stopped = False

    def stop(self):
        if not self._stopped:
            self._reset_frame_index()
            # unschedule the animation method
            pyglet.clock.unschedule(self._animate)
            # set the texture to the first frame of current animation
            frame = self._animation.frames[self._frame_index]
            self._set_texture(frame.image.get_texture())
            self._stopped = True