__author__ = 'tomlun'

import math

import pyglet
from pyglet.gl import *

import scene
import tiles
import objects
import animatedsprite
import terminalscene
import gamemenuscene
import gameoverscene


class GameScene(scene.Scene):
    _level = None
    _player = None
    _paused = False
    _interact_label = None

    def build(self):
        self._scene_manager.push_handlers(self.on_pop_scene)
        self._level = Level.load('1.level')
        self._player = Player(*self._level.get_spawn())

        self._interact_label = pyglet.text.Label(
            '[SPACE] Use',
            font_name='wendy',
            font_size=18,
            x=self._scene_manager.get_window_width() / 2 + 24,
            y=self._scene_manager.get_window_height() / 2 - 24,
            anchor_x='center', anchor_y='center',
        )

    def on_pop_scene(self):
        if self._paused:
            self._paused = False

    def update(self, dt):
        if self._paused:
            return

        self._player.update(dt)
        if not self._level.update(dt, self._player):
            # game over
            self._paused = True
            self._scene_manager.push_scene(gameoverscene.GameMenuScene(self._scene_manager))

    def draw(self):
        glLoadIdentity()
        glTranslatef(
            -self._player.x + self._scene_manager.get_window_width() / 2,
            -self._player.y + self._scene_manager.get_window_height() / 2,
            -1
        )

        self._level.draw()
        self._player.draw()

        # draw hud
        glLoadIdentity()
        if self._player.interactable_object:
            self._interact_label.draw()

    def on_key_press(self, symbol, modifiers):
        # movement
        if symbol == pyglet.window.key.UP:
            self._player.movement = Player.MOVEMENT_UP
            self._player.stop()
            self._player.play(1, 2, True)
        elif symbol == pyglet.window.key.DOWN:
            self._player.movement = Player.MOVEMENT_DOWN
            self._player.stop()
            self._player.play(1, 2, True)
        elif symbol == pyglet.window.key.LEFT:
            self._player.movement = Player.MOVEMENT_LEFT
            self._player.stop()
            self._player.play(3, 4, True)
        elif symbol == pyglet.window.key.RIGHT:
            self._player.movement = Player.MOVEMENT_RIGHT
            self._player.stop()
            self._player.play(3, 4, True)

    def on_key_release(self, symbol, modifiers):
        if symbol == pyglet.window.key.ESCAPE:
            self._paused = True
            self._scene_manager.push_scene(gamemenuscene.GameMenuScene(self._scene_manager))

        # interaction
        if symbol == pyglet.window.key.SPACE:
            if self._player.interactable_object:
                obj = self._player.interactable_object
                if isinstance(obj, objects.Door1) or isinstance(obj, objects.Door2):
                    self._level.remove_object(obj)
                elif isinstance(obj, objects.Terminal):
                    self._paused = True
                    self._scene_manager.push_scene(terminalscene.TerminalScene(self._scene_manager))
            return

        # movement
        if symbol == pyglet.window.key.UP:
            if self._player.movement == Player.MOVEMENT_UP:
                self._player.movement = Player.MOVEMENT_NONE
                self._player.stop()
        elif symbol == pyglet.window.key.DOWN:
            if self._player.movement == Player.MOVEMENT_DOWN:
                self._player.movement = Player.MOVEMENT_NONE
                self._player.stop()
        elif symbol == pyglet.window.key.LEFT:
            if self._player.movement == Player.MOVEMENT_LEFT:
                self._player.movement = Player.MOVEMENT_NONE
                self._player.stop()
        elif symbol == pyglet.window.key.RIGHT:
            if self._player.movement == Player.MOVEMENT_RIGHT:
                self._player.movement = Player.MOVEMENT_NONE
                self._player.stop()


class Level(object):

    def __init__(self, data, objs, npcs):
        self._tile_batch = pyglet.graphics.Batch()
        self._object_batch = pyglet.graphics.Batch()
        self._guard_batch = pyglet.graphics.Batch()
        self.tiles = []
        self._objects = []
        self._guards = []
        self._spawn = (0, 0)

        # reverse tile array to work with pyglet coordinates
        data.reverse()

        # create tiles from the level data
        for y in range(0, len(data)):
            for x in range(0, len(data[y])):
                tile = tiles.create_tile(data[y][x], x, y, self._tile_batch)
                try:
                    self.tiles[y].append(tile)
                except IndexError:
                    self.tiles.append([tile])
        # create objects
        for obj in objs:
            self._objects.append(
                objects.create_object(
                    obj[0],
                    obj[1],
                    obj[2],
                    self._object_batch
                )
            )
        # create npcs
        for npc in npcs:
            pos = npc[0].split(',')
            route = [list(map(int, i.split(','))) for i in npc[1].split(':')]

            self._guards.append(
                Guard(
                    int(pos[0]) * tiles.TILE_SIZE + 8,
                    int(pos[1]) * tiles.TILE_SIZE + 8,
                    self._guard_batch,
                    route,
                    self
                )
            )

    def remove_object(self, obj):
        if obj in self._objects:
            self._objects.remove(obj)

    def get_spawn(self):
        return self._spawn

    def set_spawn(self, pos):
        self._spawn = pos

    def _collides(self, a, b):
        # checks if two sprites collides
        return a.x < b.x + b.width and a.x + a.width > b.x + 1 \
            and a.y < b.y + b.height and a.y + a.height > b.y + 1

    def update(self, dt, player):
        # check collision with tiles
        for row in self.tiles:
            for tile in row:
                if self._collides(player, tile):
                    if tile.is_blocking():
                        player.set_position(*player.last_pos)
                    elif tile.is_lit():
                        return False
                tile.unlit()

        # check collisions with objects
        for obj in self._objects:
            if self._collides(player, obj):
                player.set_position(*player.last_pos)

        # check if we are close enough to interact with any objects
        player.can_interact_with(self._objects)

        # update all guards
        for guard in self._guards:
            guard.update(dt)
            guard.draw_torch()

        return True

    def draw(self):
        self._tile_batch.draw()
        self._object_batch.draw()
        self._guard_batch.draw()


    @staticmethod
    def load(filename):
        _tiles = []
        _objects = []
        _npcs = []
        _spawn = (0, 0)

        with pyglet.resource.file(filename) as f:
            ch = 0  # current header we are reading (1 = tiles, 2 = spawn, 3 = objects, 4 = npcs)
            for line in f:
                line = line.strip()
                if not line:
                    continue  # skip empty lines

                if line.startswith('#'):
                    continue  # comment, skip

                if line.startswith('-'):
                    h = line[2:]
                    if h == 'tiles':
                        ch = 1
                    elif h == 'spawn':
                        ch = 2
                    elif h == 'objects':
                        ch = 3
                    elif h == 'npcs':
                        ch = 4
                    else:
                        ch = 0
                else:
                    if ch == 1:  # tiles
                        _tiles.append(list(map(int, list(line))))
                    elif ch == 2:  # spawn
                        args = line.split(',')
                        _spawn = (
                            int(args[0]) * tiles.TILE_SIZE + 12,
                            int(args[1]) * tiles.TILE_SIZE + 12
                        )
                    elif ch == 3:  # objects
                        _objects.append(list(map(int, line.split(','))))
                    elif ch == 4:  # npcs
                        _npcs.append(line.split('|'))

        l = Level(_tiles, _objects, _npcs)
        l.set_spawn(_spawn)
        return l


class Player(animatedsprite.AnimatedSprite):

    SPEED = 70
    INTERACT_RADIUS = tiles.TILE_SIZE * 1.5
    MOVEMENT_NONE = 0
    MOVEMENT_UP = 1
    MOVEMENT_DOWN = 2
    MOVEMENT_LEFT = 3
    MOVEMENT_RIGHT = 4

    def __init__(self, x, y):
        animatedsprite.AnimatedSprite.__init__(
            self,
            pyglet.resource.texture('player.png'),
            1, 5, 0.3,
        )

        self.x = x
        self.y = y

        self.last_pos = (x, y)
        self.movement = self.MOVEMENT_NONE
        self.interactable_object = None

    def get_center_x(self):
        return self.x + self.width / 2

    def get_center_y(self):
        return self.y + self.height / 2

    def can_interact_with(self, objs):
        self.interactable_object = None
        for obj in objs:
            if not isinstance(obj, objects.InteractableObject):
                continue

            dx = obj.center_x - self.get_center_x()
            dy = obj.center_y - self.get_center_y()

            if ((dx ** 2) + (dy ** 2)) < self.INTERACT_RADIUS ** 2:
                self.interactable_object = obj

    def update(self, dt):
        if self.movement == self.MOVEMENT_NONE:
            return

        # save the last position for collision
        self.last_pos = (self.x, self.y)

        # do movement
        if self.movement == self.MOVEMENT_UP:
            self.y += self.SPEED * dt
        elif self.movement == self.MOVEMENT_DOWN:
            self.y -= self.SPEED * dt
        elif self.movement == self.MOVEMENT_LEFT:
            self.x -= self.SPEED * dt
        elif self.movement == self.MOVEMENT_RIGHT:
            self.x += self.SPEED * dt


class Guard(animatedsprite.AnimatedSprite):

    SPEED = 60
    DIRECTION_UP = 0
    DIRECTION_DOWN = 1
    DIRECTION_LEFT = 2
    DIRECTION_RIGHT = 3

    def __init__(self, x, y, batch, route, parent):
        animatedsprite.AnimatedSprite.__init__(
            self,
            pyglet.resource.texture('guard.png'),
            1, 5, 0.3,
            batch
        )

        self.x = x
        self.y = y

        self._next_route_index = 1
        self._route = route
        self._parent = parent
        self._direction = self.DIRECTION_LEFT

    def get_center_x(self):
        return self.x + self.width / 2

    def get_center_y(self):
        return self.y + self.height / 2

    def draw_torch(self):
        start = 0
        end = 0
        if self._direction == self.DIRECTION_RIGHT:
            # special case
            for degree in xrange(0, 30 + 1, 10):
                r = math.radians(degree)
                x = -self.get_center_x() + (math.cos(r) * (tiles.TILE_SIZE * 5))
                y = -self.get_center_y() + (math.cos(r) * (tiles.TILE_SIZE * 5))
                for tile in self._raycast(x, y, r):
                    tile.lit()
            for degree in xrange(330, 360 + 1, 10):
                r = math.radians(degree)
                x = -self.get_center_x() + (math.cos(r) * (tiles.TILE_SIZE * 5))
                y = -self.get_center_y() + (math.cos(r) * (tiles.TILE_SIZE * 5))
                for tile in self._raycast(x, y, r):
                    tile.lit()
            return
        if self._direction == self.DIRECTION_UP:
            start = 60
            end = 120
        if self._direction == self.DIRECTION_LEFT:
            start = 150
            end = 210
        if self._direction == self.DIRECTION_DOWN:
            start = 230
            end = 290

        for degree in xrange(start, end + 1, 10):
            r = math.radians(degree)
            x = -self.get_center_x() + (math.cos(r) * (tiles.TILE_SIZE * 5))
            y = -self.get_center_y() + (math.cos(r) * (tiles.TILE_SIZE * 5))
            for tile in self._raycast(x, y, r):
                tile.lit()

    def _raycast(self, end_x, end_y, r):
        # simple line-algorithm, probably messed up as hell since it came from my own head
        # watch out for monsters in here!
        end_tile_x = int(end_x / tiles.TILE_SIZE)
        end_tile_y = int(end_y / tiles.TILE_SIZE)

        tile_x = int(self.get_center_x() / tiles.TILE_SIZE)
        tile_y = int(self.get_center_y() / tiles.TILE_SIZE)

        if tile_x != end_tile_x:
            x = self.get_center_x() + (math.cos(r) * tiles.TILE_SIZE)
        else:
            x = end_x
        if tile_y != end_tile_y:
            y = self.get_center_y() + (math.sin(r) * tiles.TILE_SIZE)
        else:
            y = end_y

        i = 0
        while True:
            tile_x = int(x / tiles.TILE_SIZE)
            tile_y = int(y / tiles.TILE_SIZE)

            tile = self._parent.tiles[tile_y][tile_x]
            if tile.is_blocking():
                return

            # guards can only see 5 tiles away
            if i == 4:
                return

            yield tile
            if tile_x == end_tile_x and tile_y == end_tile_y:
                break

            if tile_x != end_tile_x:
                x += (math.cos(r) * tiles.TILE_SIZE)
            if tile_y != end_tile_y:
                y += (math.sin(r) * tiles.TILE_SIZE)
            i += 1

    def update(self, dt):
        n = 2
        r = self._route[self._next_route_index]
        r_x = r[0] * tiles.TILE_SIZE + 8
        r_y = r[1] * tiles.TILE_SIZE + 8
        if r_x not in range(int(self.x) - n, int(self.x) + n):
            if r_x < self.x:
                self.x -= self.SPEED * dt
                if self._direction != self.DIRECTION_LEFT:
                    self._direction = self.DIRECTION_LEFT
                    self.stop()
                    self.play(3, 4, True)
            else:
                self.x += self.SPEED * dt
                if self._direction != self.DIRECTION_RIGHT:
                    self._direction = self.DIRECTION_RIGHT
                    self.stop()
                    self.play(3, 4, True)
        if r_y not in range(int(self.y) - n, int(self.y) + n):
            if r_y < self.y:
                self.y -= self.SPEED * dt
                if self._direction != self.DIRECTION_DOWN:
                    self._direction = self.DIRECTION_DOWN
                    self.stop()
                    self.play(1, 2, True)
            else:
                self.y += self.SPEED * dt
                if self._direction != self.DIRECTION_UP:
                    self._direction = self.DIRECTION_UP
                    self.stop()
                    self.play(1, 2, True)

        if r_x in range(int(self.x) - n, int(self.x) + n) and r_y in range(int(self.y) - n, int(self.y) + n):
            self._next_route_index += 1
            if self._next_route_index >= len(self._route):
                self._next_route_index = 0